"""core runtime code for online, realtime tracking"""
from __future__ import with_statement, division
import threading, time, socket, sys, os, copy, struct
import warnings
import json
import collections

import tzlocal
import flydra_core.reconstruct
import numpy
import numpy as np
from numpy import nan
import Queue
from distutils.version import LooseVersion

pytables_filt = numpy.asarray
import atexit

import flydra_core.version
import flydra_core.kalman.flydra_kalman_utils as flydra_kalman_utils
import flydra_core.kalman.flydra_tracker

import flydra_core.data_descriptions
from flydra_core.coordinate_receiver import CoordinateProcessor, ATTEMPT_DATA_RECOVERY

# ensure that pytables uses numpy:
import tables

# bug was fixed in pytables 1.3.1 where HDF5 file kept in inconsistent state
assert LooseVersion(tables.__version__) >= LooseVersion("1.3.1")

import tables.flavor

tables.flavor.restrict_flavors(keep=["numpy"])
warnings.filterwarnings("ignore", category=tables.NaturalNameWarning)

import roslib

roslib.load_manifest("rospy")
roslib.load_manifest("std_srvs")
roslib.load_manifest("ros_flydra")
roslib.load_manifest("triggerbox")
import rospy
import std_srvs.srv
import std_msgs.msg
from ros_flydra.msg import FlydraError, CameraList
import ros_flydra.srv
import ros_flydra.cv2_bridge

from triggerbox.triggerbox_client import TriggerboxClient
from triggerbox.triggerbox_host import TriggerboxHost

import flydra_core.rosutils

LOG = flydra_core.rosutils.Log(to_ros=True)

MIN_KALMAN_OBSERVATIONS_TO_SAVE = (
    0  # how many data points are required before saving trajectory?
)

import flydra_core.common_variables
import flydra_core.flydra_socket as flydra_socket

WIRE_ORDER_CUR_VAL_IDX = flydra_core.data_descriptions.WIRE_ORDER_CUR_VAL_IDX
WIRE_ORDER_MEAN_VAL_IDX = flydra_core.data_descriptions.WIRE_ORDER_MEAN_VAL_IDX
WIRE_ORDER_SUMSQF_VAL_IDX = flydra_core.data_descriptions.WIRE_ORDER_SUMSQF_VAL_IDX


class MainBrainKeeper:
    def __init__(self):
        self.kept = []
        atexit.register(self.atexit)

    def register(self, mainbrain_instance):
        self.kept.append(mainbrain_instance)

    def atexit(self):
        for k in self.kept:
            k.quit()  # closes hdf5 file and closes cameras


main_brain_keeper = MainBrainKeeper()  # global to close MainBrain instances upon exit


class LockedValue:
    def __init__(self, initial_value=None):
        self.lock = threading.Lock()
        self._val = initial_value
        self._q = Queue.Queue()

    def set(self, value):
        self._q.put(value)

    def get(self):
        try:
            while 1:
                self._val = self._q.get_nowait()
        except Queue.Empty:
            pass
        return self._val


# 2D data format for PyTables:
Info2D = flydra_core.data_descriptions.Info2D
TextLogDescription = flydra_core.data_descriptions.TextLogDescription
CamSyncInfo = flydra_core.data_descriptions.CamSyncInfo
HostClockInfo = flydra_core.data_descriptions.HostClockInfo
TriggerClockInfo = flydra_core.data_descriptions.TriggerClockInfo
MovieInfo = flydra_core.data_descriptions.MovieInfo
ExperimentInfo = flydra_core.data_descriptions.ExperimentInfo

FilteredObservations = flydra_kalman_utils.FilteredObservations
ML_estimates_2d_idxs_type = flydra_kalman_utils.ML_estimates_2d_idxs_type

h5_obs_names = tables.Description(FilteredObservations().columns)._v_names

# allow rapid building of numpy.rec.array:
Info2DCol_description = tables.Description(Info2D().columns)._v_nested_descr


def save_ascii_matrix(filename, m):
    fd = open(filename, mode="wb")
    for row in m:
        fd.write(" ".join(map(str, row)))
        fd.write("\n")


class TimestampEchoReceiver(threading.Thread):
    def __init__(self, main_brain):
        self.main_brain = main_brain
        threading.Thread.__init__(self, name="TimestampEchoReceiver thread")

    def run(self):
        ip2hostname = {}

        timestamp_echo_fmt2 = flydra_core.common_variables.timestamp_echo_fmt2

        port = flydra_core.common_variables.timestamp_echo_gatherer_port  # my port
        addrinfo = flydra_socket.make_addrinfo(
            host=flydra_socket.get_bind_address(), port=port
        )
        timestamp_echo_gatherer = flydra_socket.FlydraTransportReceiver(addrinfo)
        addrinfo = timestamp_echo_gatherer.get_listen_addrinfo()
        LOG.info("MainBrain TimestampEchoReceiver binding %s" % (addrinfo.to_dict(),))

        last_clock_diff_measurements = collections.defaultdict(list)

        while 1:
            try:
                timestamp_echo_buf, sender_sockaddr = timestamp_echo_gatherer.recv(
                    return_sender_sockaddr=True
                )
            except Exception as err:
                LOG.warn("unknown Exception receiving timestamp echo data: %s" % err)
                continue
            except:
                LOG.warn("unknown error (non-Exception!) receiving timestamp echo data")
                continue
            (timestamp_echo_remote_ip, cam_port) = sender_sockaddr

            stop_timestamp = time.time()

            start_timestamp, remote_timestamp = struct.unpack(
                timestamp_echo_fmt2, timestamp_echo_buf
            )

            tlist = last_clock_diff_measurements[timestamp_echo_remote_ip]
            tlist.append((start_timestamp, remote_timestamp, stop_timestamp))
            if len(tlist) == 100:
                if timestamp_echo_remote_ip not in ip2hostname:
                    ip2hostname[timestamp_echo_remote_ip] = socket.getfqdn(
                        timestamp_echo_remote_ip
                    )
                remote_hostname = ip2hostname[timestamp_echo_remote_ip]
                tarray = numpy.array(tlist)

                del tlist[:]  # clear list
                start_timestamps = tarray[:, 0]
                stop_timestamps = tarray[:, 2]
                roundtrip_duration = stop_timestamps - start_timestamps
                # find best measurement (that with shortest roundtrip_duration)
                rowidx = numpy.argmin(roundtrip_duration)
                srs = tarray[rowidx, :]
                start_timestamp, remote_timestamp, stop_timestamp = srs
                clock_diff_msec = abs(remote_timestamp - start_timestamp) * 1e3
                if clock_diff_msec > 1:
                    self.main_brain.error_ros_msgs_pub.publish(
                        FlydraError(
                            FlydraError.CLOCK_DIFF,
                            "%s/%f" % (remote_hostname, clock_diff_msec),
                        )
                    )
                    LOG.warn(
                        "%s : clock diff: %.3f msec(measurement err: %.3f msec)"
                        % (
                            remote_hostname,
                            clock_diff_msec,
                            roundtrip_duration[rowidx] * 1e3,
                        )
                    )

                self.main_brain.queue_host_clock_info.put(
                    (remote_hostname, start_timestamp, remote_timestamp, stop_timestamp)
                )
                if 0:
                    measurement_duration = roundtrip_duration[rowidx]
                    clock_diff = stop_timestamp - remote_timestamp

                    LOG.debug(
                        "%s: the remote diff is %.1f msec (within 0-%.1f msec accuracy)"
                        % (
                            remote_hostname,
                            clock_diff * 1000,
                            measurement_duration * 1000,
                        )
                    )


class MainBrain(object):
    """Handle all camera network stuff and interact with application"""

    # See commits explaining socket starvation on why these are not all enabled
    ROS_CONTROL_API = dict(
        start_collecting_background=(std_srvs.srv.Empty),
        stop_collecting_background=(std_srvs.srv.Empty),
        take_background=(std_srvs.srv.Empty),
        #        clear_background=(std_srvs.srv.Empty),
        start_saving_data=(std_srvs.srv.Empty),
        stop_saving_data=(std_srvs.srv.Empty),
        start_recording=(std_srvs.srv.Empty),
        stop_recording=(std_srvs.srv.Empty),
        start_small_recording=(std_srvs.srv.Empty),
        stop_small_recording=(std_srvs.srv.Empty),
        do_synchronization=(std_srvs.srv.Empty),
        log_message=(ros_flydra.srv.MainBrainLogMessage),
        get_version=(ros_flydra.srv.MainBrainGetVersion),
        register_new_camera=(ros_flydra.srv.MainBrainRegisterNewCamera),
        get_listen_address=(ros_flydra.srv.MainBrainGetListenAddress),
        get_and_clear_commands=(ros_flydra.srv.MainBrainGetAndClearCommands),
        set_image=(ros_flydra.srv.MainBrainSetImage),
        receive_missing_data=(ros_flydra.srv.MainBrainReceiveMissingData),
        close_camera=(ros_flydra.srv.MainBrainCloseCamera),
    )

    ROS_CONFIGURATION = dict(
        frames_per_second=100.0,
        triggerbox_namespace="/trig1",
        triggerbox_hardware_device="",
        kalman_model="EKF mamarama, units: mm",
        max_reconstruction_latency_sec=0.06,  # 60 msec
        max_N_hypothesis_test=3,
        save_data_dir="~/FLYDRA",
        save_movie_dir="~/FLYDRA_MOVIES",
        camera_calibration="",
        use_unix_domain_sockets=False,
        posix_scheduler="",  # '' means OS default, set to e.g. ['FIFO', 99] for max
    )

    class RemoteAPI:

        # ================================================================
        #
        # Methods called locally
        #
        # ================================================================

        def post_init(self, main_brain):
            """call after __init__"""
            self.cam_info = {}
            self.cam_info_lock = threading.Lock()
            self.changed_cam_lock = threading.Lock()
            self.no_cams_connected = threading.Event()
            self.no_cams_connected.set()
            with self.changed_cam_lock:
                self.new_cam_ids = []
                self.old_cam_ids = []
            self.main_brain = main_brain

            # threading control locks
            self.quit_now = threading.Event()
            self.thread_done = threading.Event()
            self.message_queue = Queue.Queue()

        def external_get_and_clear_pending_cams(self):
            with self.changed_cam_lock:
                new_cam_ids = self.new_cam_ids
                self.new_cam_ids = []
                old_cam_ids = self.old_cam_ids
                self.old_cam_ids = []
            return new_cam_ids, old_cam_ids

        def external_get_cam_ids(self):
            with self.cam_info_lock:
                cam_ids = self.cam_info.keys()
            cam_ids.sort()
            return cam_ids

        def external_get_info(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    scalar_control_info = copy.deepcopy(cam["scalar_control_info"])
                    fqdn = cam["fqdn"]
                    camnode_ros_name = cam["camnode_ros_name"]
            return scalar_control_info, fqdn, camnode_ros_name

        def external_get_image_fps_points(self, cam_id):
            ### XXX should extend to include lines
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    coord_and_image = cam["image"]
                    points_distorted = cam["points_distorted"][:]
            # NB: points are distorted (and therefore align
            # with distorted image)
            if coord_and_image is not None:
                image_coords, image = coord_and_image
            else:
                image_coords, image = None, None
            fps = np.nan
            return image, fps, points_distorted, image_coords

        def external_send_set_camera_property(self, cam_id, property_name, value):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"].setdefault("set", {})[property_name] = value
                    old_value = cam["scalar_control_info"][property_name]
                    if type(old_value) == tuple and type(value) == int:
                        # brightness, gain, shutter
                        cam["scalar_control_info"][property_name] = (
                            value,
                            old_value[1],
                            old_value[2],
                        )
                    else:
                        cam["scalar_control_info"][property_name] = value

        def external_request_image_async(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["get_im"] = None

        def external_start_recording(self, cam_id, raw_file_basename):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["start_recording"] = raw_file_basename

        def external_stop_recording(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["stop_recording"] = None

        def external_start_small_recording(self, cam_id, small_filebasename):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["start_small_recording"] = small_filebasename

        def external_stop_small_recording(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["stop_small_recording"] = None

        def external_quit(self, cam_id):
            with self.cam_info_lock:
                if cam_id in self.cam_info:
                    cam = self.cam_info[cam_id]
                    cam_lock = cam["lock"]
                    with cam_lock:
                        cam["commands"]["quit"] = True

        def external_take_background(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["take_bg"] = None

        def external_request_missing_data(
            self, cam_id, camn, framenumber_offset, list_of_missing_framenumbers
        ):
            with self.cam_info_lock:
                if cam_id not in self.cam_info:
                    # the camera was dropped, ignore this request
                    return
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]

                camn_and_list = [camn, framenumber_offset]
                camn_and_list.extend(list_of_missing_framenumbers)
                cmd_str = " ".join(map(repr, camn_and_list))
                with cam_lock:
                    cam["commands"]["request_missing"] = cmd_str
            LOG.info(
                "requested missing data from %s. offset %d, frames %s"
                % (cam_id, framenumber_offset, list_of_missing_framenumbers)
            )

        def external_clear_background(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cam["commands"]["clear_bg"] = None

        # ================================================================
        #
        # Methods called remotely from cameras
        #
        # These all get called in their own thread.  Don't call across
        # the thread boundary without using locks, especially to GUI
        # or OpenGL.
        #
        # ================================================================

        def register_new_cam(
            self, cam_guid, scalar_control_info, camnode_ros_name, cam_hostname
        ):
            """register new camera (caller: remote camera)"""

            assert camnode_ros_name is not None
            fqdn = cam_hostname

            do_close = False
            with self.cam_info_lock:
                if cam_guid in self.cam_info:
                    do_close = True

            if do_close:
                LOG.warn("camera %s already exists, clearing existing data" % cam_guid)
                self.close(cam_guid)
                self.main_brain.service_pending()

            LOG.info(
                "REGISTER NEW CAMERA %s on %s @ ros node %s"
                % (cam_guid, fqdn, camnode_ros_name)
            )
            self.main_brain.coord_processor.connect(cam_guid)

            with self.cam_info_lock:
                self.cam_info[cam_guid] = {
                    "commands": {},  # command queue for cam
                    "lock": threading.Lock(),  # prevent concurrent access
                    "image": None,  # most recent image from cam
                    "points_distorted": [],  # 2D image points
                    "scalar_control_info": scalar_control_info,
                    "fqdn": fqdn,
                    "camnode_ros_name": camnode_ros_name,
                }
            self.no_cams_connected.clear()
            with self.changed_cam_lock:
                self.new_cam_ids.append(cam_guid)

        def get_listen_addr(self):
            return self.main_brain.coord_processor.get_listen_address()

        def set_image(self, cam_id, coord_and_image):
            """set most recent image (caller: remote camera)"""
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    self.cam_info[cam_id]["image"] = coord_and_image

        def receive_missing_data(self, cam_id, framenumber_offset, missing_data):
            rospy.loginfo(
                "received requested stale data for frame %d" % framenumber_offset
            )

            if len(missing_data) == 0:
                # no missing data
                return

            deferred_2d_data = []
            for (
                absolute_cam_no,
                framenumber,
                remote_timestamp,
                camn_received_time,
                points_distorted,
            ) in missing_data:

                corrected_framenumber = framenumber - framenumber_offset
                if len(points_distorted) == 0:
                    # No point was tracked that frame, send nan values.
                    points_distorted = [(nan, nan, nan, nan, nan, False, 0, 0, 0)]
                for frame_pt_idx, point_tuple in enumerate(points_distorted):
                    # Save 2D data (even when no point found) to allow
                    # temporal correlation of movie frames to 2D data.
                    try:
                        cur_val = point_tuple[WIRE_ORDER_CUR_VAL_IDX]
                        mean_val = point_tuple[WIRE_ORDER_MEAN_VAL_IDX]
                        sumsqf_val = point_tuple[WIRE_ORDER_SUMSQF_VAL_IDX]
                    except:
                        LOG.warn("error while appending point_tuple %r" % point_tuple)
                        raise
                    if corrected_framenumber is None:
                        # don't bother saving if we don't know when it was from
                        continue

                    point_tuple5 = tuple(point_tuple[:5])
                    deferred_2d_data.append(
                        (
                            absolute_cam_no,  # defer saving to later
                            corrected_framenumber,
                            remote_timestamp,
                            camn_received_time,
                        )
                        + point_tuple5
                        + (frame_pt_idx, cur_val, mean_val, sumsqf_val)
                    )
            self.main_brain.queue_data2d.put(deferred_2d_data)

        def get_and_clear_commands(self, cam_id):
            with self.cam_info_lock:
                cam = self.cam_info[cam_id]
                cam_lock = cam["lock"]
                with cam_lock:
                    cmds = cam["commands"]
                    cam["commands"] = {}
            return cmds

        def log_message(self, cam_id, timestamp, message):
            mainbrain_timestamp = time.time()
            LOG.info("received log message from %s: %s" % (cam_id, message))
            self.message_queue.put((mainbrain_timestamp, cam_id, timestamp, message))

        def close(self, cam_id):
            """gracefully say goodbye (caller: remote camera)"""
            with self.cam_info_lock:
                self.main_brain.coord_processor.disconnect(cam_id)
                del self.cam_info[cam_id]
                if not len(self.cam_info):
                    self.no_cams_connected.set()
                with self.changed_cam_lock:
                    self.old_cam_ids.append(cam_id)

    ######## end of RemoteAPI class

    # main MainBrain class

    def __init__(self, server=None, save_profiling_data=False, show_sync_errors=True):
        global main_brain_keeper

        if server is not None:
            LOG.warn("deprecated 'server' argument given.")

        LOG.info('ros node name "%s"' % rospy.get_name())

        self.load_config()

        self.debug_level = threading.Event()
        self.show_overall_latency = threading.Event()

        self._is_synchronizing = False

        # we support in or out of process trigger boxes
        if self.config["triggerbox_hardware_device"]:
            # in process
            self.trigger_device = TriggerboxHost(
                device=self.config["triggerbox_hardware_device"],
                ros_topic_base=self.config["triggerbox_namespace"],
            )
        else:
            # out of process
            self.trigger_device = TriggerboxClient(
                host_node=self.config["triggerbox_namespace"]
            )

        self.trigger_device.clock_measurement_callback = (
            self._on_trigger_clock_measurement
        )
        self.trigger_device.set_frames_per_second_blocking(
            self.config["frames_per_second"]
        )

        self.block_triggerbox_activity = False

        remote_api = MainBrain.RemoteAPI()
        remote_api.post_init(self)

        self.remote_api = remote_api

        self._config_change_functions = []
        self._new_camera_functions = []
        self._old_camera_functions = []

        self.last_requested_image = {}
        self.pending_requests = {}
        self.last_set_param_time = {}

        self.cam_host_sockets = {}

        self.num_cams = 0
        self.MainBrain_cam_ids_copy = []  # keep a copy of all cam_ids connected
        self._ip_addrs_by_cam_id = {}
        self.set_new_camera_callback(self.IncreaseCamCounter)
        self.set_new_camera_callback(self.AddTimestampEchoer)
        self.set_new_camera_callback(self.SendExpectedFPS)
        self.set_old_camera_callback(self.DecreaseCamCounter)

        self.last_saved_data_time = 0.0

        self._currently_recording_movies = {}

        # Attributes accessed by other threads (see the corresponding @property
        # get/set-ters of the attribute for locking (if any)
        self._best_realtime_data = None
        self._framenumber = 0

        self.reconstructor = None

        # Attributes which come in use when saving data occurs
        self.close_pending = False
        self._service_save_data_lock = threading.Lock()
        self.h5file = None
        self.h5filename = ""
        self.h5data2d = None
        self.h5cam_info = None
        self.h5host_clock_info = None
        self.h5trigger_clock_info = None
        self.h5movie_info = None
        self.h5exp_info = None
        self.h5textlog = None
        if 1:
            self.h5data3d_kalman_estimates = None
            self.h5data3d_ML_estimates = None
            self.h5_2d_obs = None

        # Queues of information to save
        self.queue_data2d = Queue.Queue()
        self.queue_host_clock_info = Queue.Queue()
        self.queue_trigger_clock_info = Queue.Queue()
        self.queue_data3d_best = Queue.Queue()

        self.queue_data3d_kalman_estimates = Queue.Queue()
        self.error_ros_msgs_pub = rospy.Publisher("~error", FlydraError, queue_size=100)

        self.coord_processor = CoordinateProcessor(
            self,
            save_profiling_data=save_profiling_data,
            debug_level=self.debug_level,
            show_overall_latency=self.show_overall_latency,
            show_sync_errors=show_sync_errors,
            max_reconstruction_latency_sec=self.config[
                "max_reconstruction_latency_sec"
            ],
            max_N_hypothesis_test=self.config["max_N_hypothesis_test"],
            use_unix_domain_sockets=self.config["use_unix_domain_sockets"],
            posix_scheduler=self.config["posix_scheduler"],
        )
        # self.coord_processor.setDaemon(True)
        self.coord_processor.start()

        self.timestamp_echo_receiver = TimestampEchoReceiver(self)
        self.timestamp_echo_receiver.setDaemon(True)
        self.timestamp_echo_receiver.start()

        # setup ROS
        self.pub_data_file = rospy.Publisher(
            "~data_file", std_msgs.msg.String, queue_size=0, latch=True
        )
        self.pub_data_file.publish("")
        self.pub_calib_file = rospy.Publisher(
            "~calibration", std_msgs.msg.String, queue_size=0, latch=True
        )
        self.pub_calib_file.publish("")
        self.pub_num_cams = rospy.Publisher(
            "~num_cameras", std_msgs.msg.UInt32, queue_size=0, latch=True
        )
        self.pub_num_cams.publish(0)

        self.experiment_uuid = None
        self.sub_exp_uuid = rospy.Subscriber(
            "experiment_uuid", std_msgs.msg.String, self._on_experiment_uuid
        )

        self.services = {}
        for name, srv in self.ROS_CONTROL_API.iteritems():
            self.services[name] = rospy.Service(
                "~%s" % name, srv, self._ros_generic_service_dispatch
            )

        # final config processing
        self.load_calibration(self.config["camera_calibration"])
        self.set_new_tracker(self.config["kalman_model"])
        self.set_save_data_dir(self.config["save_data_dir"])

        main_brain_keeper.register(self)

    def _on_experiment_uuid(self, msg):
        self.experiment_uuid = msg.data
        if self.is_saving_data():
            self.h5exp_info.row["uuid"] = self.experiment_uuid
            self.h5exp_info.row.append()
            self.h5exp_info.flush()

    def _on_trigger_clock_measurement(
        self, start_timestamp, pulsenumber, fraction_n_of_255, stop_timestamp
    ):
        self.queue_trigger_clock_info.put(
            (start_timestamp, pulsenumber, fraction_n_of_255, stop_timestamp)
        )

    def _ros_generic_service_dispatch(self, req):
        calledservice = req._connection_header["service"]
        calledfunction = calledservice.split("/")[-1]
        if calledfunction in self.ROS_CONTROL_API:
            srvclass = self.ROS_CONTROL_API[calledfunction]

            # dynamically build the request and response argument lists for the mainbrain api
            # call. This requires the mainbrain api have the same calling signature as the
            # service definitions, and, if you want to return something over ros, the return
            # type signature must again match the service return type signature

            respclass = srvclass._response_class

            # determine the args to pass to the function based on the srv description (which
            # is embodied in __slots__, a variable created by the ros build system to list
            # the attributes (i.e. parameters only) of the request
            kwargs = {}
            for attr in req.__slots__:
                kwargs[attr] = getattr(req, attr)

            result = getattr(self, calledfunction)(**kwargs)

            kwargs = {}
            for i, attr in enumerate(respclass.__slots__):
                kwargs[attr] = result[i]

            return respclass(**kwargs)

    @property
    def framenumber(self):
        return self._framenumber

    @framenumber.setter
    def framenumber(self, value):
        self._framenumber = value

    @property
    def best_realtime_data(self):
        data = self._best_realtime_data
        self._best_realtime_data = None
        return data

    @best_realtime_data.setter
    def best_realtime_data(self, value):
        self._best_realtime_data = value

    def load_config(self):
        self.config = {}
        for k, v in self.ROS_CONFIGURATION.iteritems():
            self.config[k] = rospy.get_param("~%s" % k, v)

    def save_config(self):
        for k, v in self.config.iteritems():
            if k in self.ROS_CONFIGURATION:
                rospy.set_param("~%s" % k, v)
        for func in self._config_change_functions:
            func()

    def get_fps(self):
        return self.trigger_device.get_frames_per_second()

    def set_fps(self, fps):
        self.do_synchronization(new_fps=fps)

    def get_version(self):
        return (std_msgs.msg.String(flydra_core.version.__version__),)

    def log_message(self, cam_id, timestamp, message):
        self.remote_api.log_message(cam_id.data, timestamp.data, message.data)

    def register_new_camera(
        self, cam_guid, scalar_control_info_json, camnode_ros_name, cam_hostname, cam_ip
    ):
        if len(cam_ip.data) > 0:
            LOG.warn("'cam_ip' parameter set, even though it is deprecated")
        scalar_control_info = json.loads(scalar_control_info_json.data)
        self.remote_api.register_new_cam(
            cam_guid=cam_guid.data,
            scalar_control_info=scalar_control_info,
            camnode_ros_name=camnode_ros_name.data,
            cam_hostname=cam_hostname.data,
        )
        return [std_msgs.msg.Int32(-1)]

    def get_listen_address(self):
        listen_addr = self.remote_api.get_listen_addr()
        listen_addr_json = json.dumps(listen_addr)
        return (std_msgs.msg.String(listen_addr_json),)

    def get_and_clear_commands(self, cam_id):
        cmds = self.remote_api.get_and_clear_commands(cam_id.data)
        cmds_json = json.dumps(cmds)
        return [std_msgs.msg.String(cmds_json)]

    def set_image(self, cam_id, left, bottom, image):
        cam_id = cam_id.data
        lb = left.data, bottom.data
        image = ros_flydra.cv2_bridge.imgmsg_to_numpy(image)
        self.remote_api.set_image(cam_id, (lb, image))

    def receive_missing_data(self, cam_id, framenumber_offset, missing_data_json_buf):
        missing_data = json.loads(missing_data_json_buf.data)
        self.remote_api.receive_missing_data(
            cam_id.data, framenumber_offset.data, missing_data
        )

    def close_xcamera(self, cam_id):
        cam_id = cam_id.data
        self.remote_api.close(cam_id)

    def do_synchronization(self, new_fps=None):
        if self.is_saving_data():
            raise RuntimeError("will not (re)synchronize while saving data")

        self.coord_processor.delete_list_of_synced_cameras()
        self._is_synchronizing = True

        assert self.block_triggerbox_activity == False

        if new_fps is not None:
            self.trigger_device.set_frames_per_second(new_fps)
            actual_new_fps = self.trigger_device.get_frames_per_second()

            # the tracker depends on the framerate
            self.update_tracker_fps(actual_new_fps)

        self.coord_processor.mainbrain_is_attempting_synchronizing()
        self.trigger_device.synchronize(
            flydra_core.common_variables.sync_duration + 1.0
        )
        if new_fps is not None:
            cam_ids = self.remote_api.external_get_cam_ids()
            for cam_id in cam_ids:
                try:
                    self.send_set_camera_property(
                        cam_id, "expected_trigger_framerate", actual_new_fps
                    )
                except Exception as err:
                    LOG.warn("set_camera_property_error %s" % err)

            self.config["frames_per_second"] = float(actual_new_fps)
            self.save_config()

    def IncreaseCamCounter(self, cam_id, scalar_control_info, fqdn):
        self.num_cams += 1
        self.MainBrain_cam_ids_copy.append(cam_id)
        self.pub_num_cams.publish(self.num_cams)

    def AddTimestampEchoer(self, cam_id, scalar_control_info, fqdn):
        if fqdn not in self.cam_host_sockets:
            port = flydra_core.common_variables.timestamp_echo_listener_port
            addrinfo = flydra_socket.make_addrinfo(host=fqdn, port=port)
            self.cam_host_sockets[fqdn] = flydra_socket.FlydraTransportSender(addrinfo)

    def SendExpectedFPS(self, cam_id, scalar_control_info, fqdn):
        self.send_set_camera_property(
            cam_id,
            "expected_trigger_framerate",
            self.trigger_device.get_frames_per_second(),
        )

    def DecreaseCamCounter(self, cam_id):
        try:
            idx = self.MainBrain_cam_ids_copy.index(cam_id)
        except ValueError:
            LOG.warn(
                "IGNORING ERROR: DecreaseCamCounter() called with non-existant cam_id"
            )
            return
        self.num_cams -= 1
        del self.MainBrain_cam_ids_copy[idx]
        self.pub_num_cams.publish(self.num_cams)

    def get_num_cams(self):
        return self.num_cams

    def get_scalarcontrolinfo(self, cam_id):
        sci, fqdn, camnode_ros_name = self.remote_api.external_get_info(cam_id)
        return sci

    def get_widthheight(self, cam_id):
        sci, fqdn, camnode_ros_name = self.remote_api.external_get_info(cam_id)
        w = sci["width"]
        h = sci["height"]
        return w, h

    def get_roi(self, cam_id):
        sci, fqdn, camnode_ros_name = self.remote_api.external_get_info(cam_id)
        lbrt = sci["roi"]
        return lbrt

    def get_all_params(self):
        cam_ids = self.remote_api.external_get_cam_ids()
        all = {}
        for cam_id in cam_ids:
            sci, fqdn, camnode_ros_name = self.remote_api.external_get_info(cam_id)
            all[cam_id] = sci
        return all

    def start_listening(self):
        """ the last thing called before we work - give the config callback watchers a callback
        to check on the state of the mainbrain post __init__ """
        self.save_config()

    def set_config_change_callback(self, handler):
        self._config_change_functions.append(handler)

    def set_new_camera_callback(self, handler):
        self._new_camera_functions.append(handler)

    def set_old_camera_callback(self, handler):
        self._old_camera_functions.append(handler)

    def service_pending(self):
        """the MainBrain application calls this fairly frequently (e.g. every 100 msec)"""
        new_cam_ids, old_cam_ids = self.remote_api.external_get_and_clear_pending_cams()
        for cam_id in new_cam_ids:
            if cam_id in old_cam_ids:
                continue  # inserted and then removed
            if self.is_saving_data():
                raise RuntimeError("Cannot add new camera while saving data")
            sci, fqdn, camnode_ros_name = self.remote_api.external_get_info(cam_id)
            for new_cam_func in self._new_camera_functions:
                new_cam_func(cam_id, sci, fqdn)

        for cam_id in old_cam_ids:
            for old_cam_func in self._old_camera_functions:
                old_cam_func(cam_id)

        now = time.time()
        diff = now - self.last_saved_data_time
        if diff >= 5.0:  # request missing data and save data every 5 seconds
            self._request_missing_data()
            self._locked_service_save_data()
            self.last_saved_data_time = now

        self._check_latencies()

    def _check_latencies(self):
        timestamp_echo_fmt1 = flydra_core.common_variables.timestamp_echo_fmt1
        for sock in self.cam_host_sockets.itervalues():
            buf = struct.pack(timestamp_echo_fmt1, time.time())
            sock.send(buf)

    def get_last_image_fps(self, cam_id):
        # XXX should extend to include lines

        # Points are originally distorted (and align with distorted
        # image).
        (
            image,
            fps,
            points_distorted,
            image_coords,
        ) = self.remote_api.external_get_image_fps_points(cam_id)

        return image, fps, points_distorted, image_coords

    def close_camera(self, cam_id):
        sys.stdout.flush()
        self.remote_api.external_quit(cam_id)
        sys.stdout.flush()

    def start_collecting_background(self, *cam_ids):
        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            self.set_collecting_background(cam_id, True)

    def stop_collecting_background(self, *cam_ids):
        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            self.set_collecting_background(cam_id, False)

    def set_collecting_background(self, cam_id, value):
        self.remote_api.external_send_set_camera_property(
            cam_id, "collecting_background", value
        )

    def set_color_filter(self, cam_id, value):
        self.remote_api.external_send_set_camera_property(cam_id, "color_filter", value)

    def take_background(self, *cam_ids):
        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            self.remote_api.external_take_background(cam_id)

    def clear_background(self, *cam_ids):
        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            self.remote_api.external_clear_background(cam_id)

    def send_set_camera_property(self, cam_id, property_name, value):
        self.remote_api.external_send_set_camera_property(cam_id, property_name, value)

    def request_image_async(self, cam_id):
        self.remote_api.external_request_image_async(cam_id)

    def get_debug_level(self):
        return self.debug_level.isSet()

    def set_debug_level(self, value):
        if value:
            self.debug_level.set()
        else:
            self.debug_level.clear()

    def get_show_overall_latency(self):
        return self.show_overall_latency.isSet()

    def set_show_overall_latency(self, value):
        if value:
            self.show_overall_latency.set()
        else:
            self.show_overall_latency.clear()

    def start_recording(self, raw_file_basename=None, *cam_ids):
        nowstr = time.strftime("%Y%m%d_%H%M%S")
        if not raw_file_basename:
            if self.experiment_uuid is not None:
                raw_file_basename = os.path.join(
                    self.config["save_movie_dir"], self.experiment_uuid,
                )
            else:
                raw_file_basename = os.path.join(self.config["save_movie_dir"], nowstr,)

        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            raw_file_name = os.path.join(raw_file_basename, cam_id, nowstr)
            self.remote_api.external_start_recording(cam_id, raw_file_name)
            approx_start_frame = self.framenumber
            self._currently_recording_movies[cam_id] = (
                raw_file_name,
                approx_start_frame,
            )
            if self.is_saving_data():
                self.h5movie_info.row["cam_id"] = cam_id
                self.h5movie_info.row["filename"] = raw_file_name + ".fmf"
                self.h5movie_info.row["approx_start_frame"] = approx_start_frame
                self.h5movie_info.row.append()
                self.h5movie_info.flush()

    def stop_recording(self, *cam_ids):
        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            self.remote_api.external_stop_recording(cam_id)
            if cam_id not in self._currently_recording_movies:
                # we're not actually saving...
                continue
            approx_stop_frame = self.framenumber
            raw_file_basename, approx_start_frame = self._currently_recording_movies[
                cam_id
            ]
            del self._currently_recording_movies[cam_id]
            # modify save file to include approximate movie stop time
            if self.is_saving_data():
                nrow = None
                for r in self.h5movie_info:
                    # get row in table
                    if (
                        r["cam_id"] == cam_id
                        and r["filename"] == raw_file_basename + ".fmf"
                        and r["approx_start_frame"] == approx_start_frame
                    ):
                        nrow = r.nrow
                        break
                if nrow is not None:
                    nrowi = int(nrow)  # pytables bug workaround...
                    assert nrowi == nrow  # pytables bug workaround...
                    approx_stop_framei = int(approx_stop_frame)
                    assert approx_stop_framei == approx_stop_frame

                    new_columns = numpy.rec.fromarrays(
                        [[approx_stop_framei]], formats="i8"
                    )
                    self.h5movie_info.modify_columns(
                        start=nrowi, columns=new_columns, names=["approx_stop_frame"]
                    )
                else:
                    raise RuntimeError("could not find row to save movie stop frame.")

    def start_small_recording(self, raw_file_basename=None, *cam_ids):
        nowstr = time.strftime("%Y%m%d_%H%M%S")
        if not raw_file_basename:
            if self.experiment_uuid is not None:
                raw_file_basename = os.path.join(
                    self.config["save_movie_dir"], self.experiment_uuid,
                )
            else:
                raw_file_basename = os.path.join(self.config["save_movie_dir"], nowstr,)

        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            raw_file_name = os.path.join(raw_file_basename, cam_id, nowstr)
            self.remote_api.external_start_small_recording(cam_id, raw_file_name)

    def stop_small_recording(self, *cam_ids):
        if len(cam_ids) == 0:
            cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            self.remote_api.external_stop_small_recording(cam_id)

    def quit(self):
        """closes any files being saved and closes camera connections"""
        # XXX ====== non-isolated calls to remote_api being done ======
        # this may be called twice: once explicitly and once by __del__
        with self.remote_api.cam_info_lock:
            cam_ids = self.remote_api.cam_info.keys()

        for cam_id in cam_ids:
            self.close_camera(cam_id)
        self.remote_api.no_cams_connected.wait(2.0)
        self.remote_api.quit_now.set()  # tell thread to finish
        self.remote_api.thread_done.wait(0.5)  # wait for thread to finish
        if not self.remote_api.no_cams_connected.isSet():
            cam_ids = self.remote_api.cam_info.keys()
            LOG.warn("cameras failed to quit cleanly: %s" % cam_ids)
            # raise RuntimeError('cameras failed to quit cleanly: %s'%str(cam_ids))

        self.stop_saving_data()
        self.coord_processor.quit()

    def load_calibration(self, dirname):
        if self.is_saving_data():
            raise RuntimeError("Cannot (re)load calibration while saving data")

        if not dirname:
            return

        dirname = flydra_core.rosutils.decode_url(dirname)
        if os.path.exists(dirname):
            connected_cam_ids = self.remote_api.external_get_cam_ids()
            self.reconstructor = flydra_core.reconstruct.Reconstructor(dirname)
            calib_cam_ids = self.reconstructor.get_cam_ids()

            calib_cam_ids = calib_cam_ids

            self.coord_processor.set_reconstructor(self.reconstructor)

            self.pub_calib_file.publish(dirname)
            self.config["camera_calibration"] = dirname
            self.save_config()
        else:
            raise ValueError(
                "you specified loading calibration from %r, but that path does not exist"
                % dirname
            )

    def clear_calibration(self):
        if self.is_saving_data():
            raise RuntimeError("Cannot unload calibration while saving data")
        cam_ids = self.remote_api.external_get_cam_ids()
        self.reconstructor = None

        self.coord_processor.set_reconstructor(self.reconstructor)

        self.save_config()

    def update_tracker_fps(self, fps):
        self.set_new_tracker(self.dynamic_model_name, fps)

    def set_new_tracker(self, kalman_model_name, new_fps=None):
        if self.is_saving_data():
            raise RuntimeError("will not set Kalman parameters while saving data")

        self.dynamic_model_name = kalman_model_name

        if self.reconstructor is None:
            return

        fps = self.get_fps() if new_fps is None else new_fps
        dt = 1.0 / fps

        LOG.info("setting model to %s (fps: %s)" % (kalman_model_name, fps))

        dynamic_model = flydra_core.kalman.dynamic_models.get_kalman_model(
            name=kalman_model_name, dt=dt
        )

        self.kalman_saver_info_instance = flydra_kalman_utils.KalmanSaveInfo(
            name=kalman_model_name
        )
        self.KalmanEstimatesDescription = (
            self.kalman_saver_info_instance.get_description()
        )
        self.dynamic_model = dynamic_model

        self.h5_xhat_names = tables.Description(
            self.KalmanEstimatesDescription().columns
        )._v_names

        # send params over to realtime coords thread
        self.coord_processor.set_new_tracker(kalman_model=dynamic_model)
        self.coord_processor.tracker.clear_flushed_callbacks()
        self.coord_processor.tracker.set_flushed_callback(self.finally_close_save_files)

        self.config["kalman_model"] = kalman_model_name
        self.save_config()

    def __del__(self):
        self.quit()

    def _safe_makedir(self, path):
        """ raises OSError if path cannot be made """
        if not os.path.exists(path):
            os.makedirs(path)
            return path

    def set_save_data_dir(self, path):
        path = flydra_core.rosutils.decode_url(path)
        if os.path.isdir(path):
            save_data_dir = path
        else:
            try:
                save_data_dir = self._safe_makedir(path)
            except OSError:
                return None
        self.config["save_data_dir"] = save_data_dir
        self.save_config()
        LOG.info("saving data to %s" % save_data_dir)
        return save_data_dir

    def is_saving_data(self):
        return self.h5file is not None

    def start_saving_data(self, filename=None):
        if self.is_saving_data():
            return

        if not filename:
            filename = time.strftime("%Y%m%d_%H%M%S.mainbrain.h5")
        filename = os.path.join(self.config["save_data_dir"], filename)

        if os.path.exists(filename):
            raise RuntimeError("will not overwrite data file")

        self.h5filename = filename

        LOG.info("saving data to %s" % self.h5filename)
        self.pub_data_file.publish(self.h5filename)

        self.block_triggerbox_activity = True
        self.h5file = tables.open_file(
            os.path.expanduser(self.h5filename), mode="w", title="Flydra data file"
        )
        expected_rows = int(1e6)
        ct = self.h5file.create_table  # shorthand
        root = self.h5file.root  # shorthand
        self.h5data2d = ct(
            root, "data2d_distorted", Info2D, "2d data", expectedrows=expected_rows * 5
        )
        self.h5cam_info = ct(
            root, "cam_info", CamSyncInfo, "Cam Sync Info", expectedrows=500
        )
        self.h5host_clock_info = ct(
            root,
            "host_clock_info",
            HostClockInfo,
            "Host Clock Info",
            expectedrows=6 * 60 * 24,
        )  # 24 hours at 10 sec sample intervals
        self.h5trigger_clock_info = ct(
            root,
            "trigger_clock_info",
            TriggerClockInfo,
            "Trigger Clock Info",
            expectedrows=6 * 60 * 24,
        )  # 24 hours at 10 sec sample intervals
        self.h5movie_info = ct(
            root, "movie_info", MovieInfo, "Movie Info", expectedrows=500
        )
        self.h5textlog = ct(root, "textlog", TextLogDescription, "text log")
        self.h5exp_info = ct(
            root, "experiment_info", ExperimentInfo, "ExperimentInfo", expectedrows=100
        )

        self._startup_message()
        if self.reconstructor is not None:
            self.reconstructor.save_to_h5file(self.h5file)
            if 1:
                self.h5data3d_kalman_estimates = ct(
                    root,
                    "kalman_estimates",
                    self.KalmanEstimatesDescription,
                    "3d data (from Kalman filter)",
                    expectedrows=expected_rows,
                )
                self.h5data3d_kalman_estimates.attrs.dynamic_model_name = (
                    self.dynamic_model_name
                )
                self.h5data3d_kalman_estimates.attrs.dynamic_model = self.dynamic_model

                self.h5data3d_ML_estimates = ct(
                    root,
                    "ML_estimates",
                    FilteredObservations,
                    "dynamics-free maximum liklihood estimates",
                    expectedrows=expected_rows,
                )
                self.h5_2d_obs = self.h5file.create_vlarray(
                    self.h5file.root,
                    "ML_estimates_2d_idxs",
                    ML_estimates_2d_idxs_type(),  # dtype should match with tro.observations_2d
                    "camns and idxs",
                )
                self.h5_2d_obs_next_idx = 0

        general_save_info = self.coord_processor.get_general_cam_info()
        for cam_id, dd in general_save_info.iteritems():
            self.h5cam_info.row["cam_id"] = cam_id
            self.h5cam_info.row["camn"] = dd["absolute_cam_no"]
            with self.remote_api.cam_info_lock:
                self.h5cam_info.row["hostname"] = self.remote_api.cam_info[cam_id][
                    "fqdn"
                ]
            self.h5cam_info.row.append()
        self.h5cam_info.flush()

        # save raw image from each camera
        img = self.h5file.create_group(root, "images", "sample images")
        cam_ids = self.remote_api.external_get_cam_ids()
        for cam_id in cam_ids:
            image, fps, points_distorted, image_coords = self.get_last_image_fps(cam_id)
            if image is None:
                raise ValueError("image cannot be None")
            self.h5file.create_array(
                img, cam_id, image, "sample image from %s" % cam_id
            )

        self.save_config()

        if self.coord_processor.tracker is not None:
            # force all new tracked objects by killing existing tracks
            self.coord_processor.tracker.kill_all_trackers()

    def stop_saving_data(self):
        LOG.info("received request to stop saving file")
        self.close_pending = True
        if self.coord_processor.tracker is not None:
            # eventually this will trigger a call to self.finally_close_save_files()
            self.coord_processor.tracker.kill_all_trackers()
        else:
            self.finally_close_save_files()

    def finally_close_save_files(self):
        if not self.close_pending:
            return
        self.close_pending = False  # after the following, we will already be closed...

        with self._service_save_data_lock:
            LOG.info("entering final save data service call")
            self._service_save_data()  # we absolutely want to save
            LOG.info("entering done with final save data service call")
            if self.is_saving_data():
                self.h5file.close()
                self.h5file = None
                self.h5filename = ""
                self.pub_data_file.publish(self.h5filename)
                self.block_triggerbox_activity = False
                LOG.info("closed h5 file")
            else:
                LOG.info("saving already stopped, cannot stop again")
            self.h5data2d = None
            self.h5cam_info = None
            self.h5host_clock_info = None
            self.h5trigger_clock_info = None
            self.h5movie_info = None
            self.h5exp_info = None
            self.h5textlog = None
            self.h5data3d_kalman_estimates = None
            self.h5data3d_ML_estimates = None
            self.h5_2d_obs = None

            self.save_config()

    def _startup_message(self):
        textlog_row = self.h5textlog.row
        cam_id = "mainbrain"
        timestamp = time.time()

        # Get local timezone name. See https://stackoverflow.com/a/17365806/1633026
        local_tz_name = tzlocal.get_localzone()

        # This line is important (including the formatting). It is
        # read by flydra_analysis.a2.check_atmel_clock.

        list_of_textlog_data = [
            (
                timestamp,
                cam_id,
                timestamp,
                "MainBrain running at %s fps, (flydra_version %s, time_tzname0 %s)"
                % (
                    self.trigger_device.get_frames_per_second(),
                    flydra_core.version.__version__,
                    local_tz_name,
                ),
            ),
            (
                timestamp,
                cam_id,
                timestamp,
                "using flydra version %s" % (flydra_core.version.__version__,),
            ),
        ]

        list_of_textlog_data.append(
            (timestamp, cam_id, timestamp, "using numpy version %s" % numpy.__version__)
        )

        list_of_textlog_data.append(
            (
                timestamp,
                cam_id,
                timestamp,
                "using pytables version %s" % tables.__version__,
            )
        )

        for lib in ("hdf5", "zlib", "lzo", "bzip2", "blosc"):
            try:
                _, ver, _ = tables.which_lib_version(lib)
                list_of_textlog_data.append(
                    (
                        timestamp,
                        cam_id,
                        timestamp,
                        "using pytables:%s version %s" % (lib, ver),
                    )
                )
            except ValueError:
                # unknown lib
                pass

        for textlog_data in list_of_textlog_data:
            (mainbrain_timestamp, cam_id, host_timestamp, message) = textlog_data
            textlog_row["mainbrain_timestamp"] = mainbrain_timestamp
            textlog_row["cam_id"] = cam_id
            textlog_row["host_timestamp"] = host_timestamp
            textlog_row["message"] = message
            textlog_row.append()

        self.h5textlog.flush()

    def _request_missing_data(self):
        if ATTEMPT_DATA_RECOVERY:
            # request from camera computers any data that we're missing
            missing_data_dict = self.coord_processor.get_missing_data_dict()
            for (
                camn,
                (cam_id, framenumber_offset, list_of_missing_framenumbers),
            ) in missing_data_dict.iteritems():
                self.remote_api.external_request_missing_data(
                    cam_id, camn, framenumber_offset, list_of_missing_framenumbers
                )

    def _locked_service_save_data(self):
        with self._service_save_data_lock:
            self._service_save_data()

    def _service_save_data(self):
        # ** 2d data **
        #   clear queue
        list_of_rows_of_data2d = []
        try:
            while True:
                tmp = self.queue_data2d.get(0)
                list_of_rows_of_data2d.extend(tmp)
        except Queue.Empty:
            pass
        #   save
        if self.h5data2d is not None and len(list_of_rows_of_data2d):
            # it's much faster to convert to numpy first:
            recarray = numpy.rec.array(
                list_of_rows_of_data2d, dtype=Info2DCol_description
            )
            self.h5data2d.append(recarray)
            self.h5data2d.flush()

        # ** textlog **
        if self.h5textlog is not None:
            # we don't want to miss messages, so wait until we are saving
            list_of_textlog_data = []
            try:
                while True:
                    tmp = self.remote_api.message_queue.get(0)
                    list_of_textlog_data.append(tmp)
            except Queue.Empty:
                pass

            if list_of_textlog_data:
                textlog_row = self.h5textlog.row
                for textlog_data in list_of_textlog_data:
                    (
                        mainbrain_timestamp,
                        cam_id,
                        host_timestamp,
                        message,
                    ) = textlog_data
                    textlog_row["mainbrain_timestamp"] = mainbrain_timestamp
                    textlog_row["cam_id"] = cam_id
                    textlog_row["host_timestamp"] = host_timestamp
                    textlog_row["message"] = message
                    textlog_row.append()

                self.h5textlog.flush()

        if 1:
            # ** 3d data - kalman **
            q = self.queue_data3d_kalman_estimates

            #   clear queue
            list_of_3d_data = []
            try:
                while True:
                    list_of_3d_data.append(q.get(0))
            except Queue.Empty:
                pass
            if self.h5data3d_kalman_estimates is not None:
                for (
                    obj_id,
                    tro_frames,
                    tro_xhats,
                    tro_Ps,
                    tro_timestamps,
                    obs_frames,
                    obs_data,
                    observations_2d,
                    obs_Lcoords,
                ) in list_of_3d_data:

                    if len(obs_frames) < MIN_KALMAN_OBSERVATIONS_TO_SAVE:
                        # only save data with at least N observations
                        continue

                    # save observation 2d data indexes
                    this_idxs = []
                    for camns_and_idxs in observations_2d:
                        this_idxs.append(self.h5_2d_obs_next_idx)
                        self.h5_2d_obs.append(camns_and_idxs)
                        self.h5_2d_obs_next_idx += 1
                    self.h5_2d_obs.flush()

                    this_idxs = numpy.array(
                        this_idxs, dtype=numpy.uint64
                    )  # becomes obs_2d_idx (index into 'ML_estimates_2d_idxs')

                    # save observations
                    observations_frames = numpy.array(obs_frames, dtype=numpy.uint64)
                    obj_id_array = numpy.empty(
                        observations_frames.shape, dtype=numpy.uint32
                    )
                    obj_id_array.fill(obj_id)
                    observations_data = numpy.array(obs_data, dtype=numpy.float32)
                    observations_Lcoords = numpy.array(obs_Lcoords, dtype=numpy.float32)
                    list_of_obs = [
                        observations_data[:, i]
                        for i in range(observations_data.shape[1])
                    ]
                    list_of_lines = [
                        observations_Lcoords[:, i]
                        for i in range(observations_Lcoords.shape[1])
                    ]
                    array_list = (
                        [obj_id_array, observations_frames]
                        + list_of_obs
                        + [this_idxs]
                        + list_of_lines
                    )
                    obs_recarray = numpy.rec.fromarrays(array_list, names=h5_obs_names)

                    self.h5data3d_ML_estimates.append(obs_recarray)
                    self.h5data3d_ML_estimates.flush()

                    # save xhat info (kalman estimates)
                    frames = numpy.array(tro_frames, dtype=numpy.uint64)
                    timestamps = numpy.array(tro_timestamps, dtype=numpy.float64)
                    xhat_data = numpy.array(tro_xhats, dtype=numpy.float32)
                    P_data_full = numpy.array(tro_Ps, dtype=numpy.float32)
                    obj_id_array = numpy.empty(frames.shape, dtype=numpy.uint32)
                    obj_id_array.fill(obj_id)
                    list_of_xhats = [xhat_data[:, i] for i in range(xhat_data.shape[1])]
                    ksii = self.kalman_saver_info_instance
                    list_of_Ps = ksii.covar_mats_to_covar_entries(P_data_full)
                    xhats_recarray = numpy.rec.fromarrays(
                        [obj_id_array, frames, timestamps] + list_of_xhats + list_of_Ps,
                        names=self.h5_xhat_names,
                    )

                    self.h5data3d_kalman_estimates.append(xhats_recarray)
                    self.h5data3d_kalman_estimates.flush()

        # ** camera info **
        #   clear queue
        list_of_host_clock_info = []
        try:
            while True:
                list_of_host_clock_info.append(self.queue_host_clock_info.get(0))
        except Queue.Empty:
            pass
        #   save
        if self.h5host_clock_info is not None:
            host_clock_info_row = self.h5host_clock_info.row
            for host_clock_info in list_of_host_clock_info:
                (
                    remote_hostname,
                    start_timestamp,
                    remote_timestamp,
                    stop_timestamp,
                ) = host_clock_info
                host_clock_info_row["remote_hostname"] = remote_hostname
                host_clock_info_row["start_timestamp"] = start_timestamp
                host_clock_info_row["remote_timestamp"] = remote_timestamp
                host_clock_info_row["stop_timestamp"] = stop_timestamp
                host_clock_info_row.append()

            self.h5host_clock_info.flush()

        #   clear queue
        list_of_trigger_clock_info = []
        try:
            while True:
                list_of_trigger_clock_info.append(self.queue_trigger_clock_info.get(0))
        except Queue.Empty:
            pass
        #   save
        if self.h5trigger_clock_info is not None:
            row = self.h5trigger_clock_info.row
            for trigger_clock_info in list_of_trigger_clock_info:
                start_timestamp, framecount, tcnt, stop_timestamp = trigger_clock_info
                row["start_timestamp"] = start_timestamp
                row["framecount"] = framecount
                row["tcnt"] = tcnt
                row["stop_timestamp"] = stop_timestamp
                row.append()

            self.h5trigger_clock_info.flush()



#emacs, this is -*-Python-*- mode
#cython: language_level=2
import numpy as np
cimport numpy as np

import flydra_core.water as water
import flydra_core.reconstruct as reconstruct
cimport _refraction
import _refraction
from libc.math cimport atan2, sqrt, cos, sin
import _pmat_jacobian

def make_PinholeCameraWaterModelWithJacobian(*args,**kw):
    return PinholeCameraWaterModelWithJacobian(*args,**kw)

cdef class PinholeCameraWaterModelWithJacobian(_pmat_jacobian.PinholeCameraModelWithJacobian):
    def __init__(self,P,wateri,roots3and4_eps):
        self.pmat = P
        self.wateri = wateri
        assert isinstance(wateri, water.WaterInterface)
        self.n1 = wateri.n1
        self.n2 = wateri.n2
        self.roots3and4_eps = roots3and4_eps

        C = reconstruct.pmat2cam_center(self.pmat)
        C = C[:,0]
        self.camx, self.camy, self.camz = C

        self.shift = np.zeros((3,))
        self.shift[0] = self.camx
        self.shift[1] = self.camy

        self.delta=0.001

        self.dx = np.array( (self.delta, 0,          0) )
        self.dy = np.array( (0,          self.delta, 0) )

        self.pinhole = _pmat_jacobian.PinholeCameraModelWithJacobian(self.pmat)

    def __reduce__(self):
        """this allows PinholeCameraWaterModelWithJacobian to be pickled"""
        args = (self.pmat,self.wateri)
        return (make_PinholeCameraWaterModelWithJacobian, args)

    cpdef evaluate_jacobian_at(self, np.ndarray[np.double_t, ndim=1] X):
        cdef int flip_z
        cdef np.ndarray[np.double_t, ndim=1] eval_z
        cdef np.ndarray[np.double_t, ndim=1] dz
        cdef double z_delta

        # evaluate the Jacobian numerically
        F = self.evaluate(X)

        Fx = self.evaluate(X + self.dx)
        Fy = self.evaluate(X + self.dy)
        z_delta = min( abs(X[2]), self.delta ) # do not allow z>0
        dz = np.array( (0, 0, z_delta) )

        eval_z = X+dz
        if eval_z[2] > 0:
            flip_z = False
        else:
            flip_z = True
            eval_z = X-dz
        Fz = self.evaluate(eval_z)

        dF_dx = (Fx-F)/self.delta
        dF_dy = (Fy-F)/self.delta
        dF_dz = (Fz-F)/self.delta
        if flip_z:
            dF_dz = -dF_dz

        result = np.array([dF_dx,
                           dF_dy,
                           dF_dz]).T

        return result

    cpdef evaluate(self, np.ndarray[np.double_t, ndim=1] X):
        """evaluate the non-linear function

        Y = h(X)
        """
        cdef double depth, r, theta, r0
        cdef double s1, s0

        cdef epsilon = 1e-16
        cdef scale=1.0
        cdef np.ndarray[np.double_t, ndim=1] surfX

        assert len(X)==3

        pt_shifted = X-self.shift
        s0,s1,depth = pt_shifted
        depth = -depth # convert from -z to z

        if depth < 0:
            # do not use refraction model if point above water
            return self.pinhole.evaluate( X )

        theta = atan2( s1, s0 )
        r = sqrt( s0*s0 + s1*s1 )
        r0 = _refraction.find_fastest_path_fermat( self.n1, self.n2, self.camz,
                                                   r, depth, self.roots3and4_eps )

        t0 = r0*cos(theta)
        t1 = r0*sin(theta)

        surfX_shifted = np.array((t0,t1,0),dtype=np.float)
        surfX = surfX_shifted+self.shift
        result = self.pinhole.evaluate(surfX)
        return result



import numpy as np
import flydra_core._refraction as _refraction

R2D = 180.0 / np.pi


class WaterInterface:
    """for now empty

    If water is ever not at z=0, this will be the place to implement it.
    """

    def __init__(self, refractive_index, water_roots_eps):
        # values for air and water from
        # http://en.wikipedia.org/wiki/Refraction
        self.n1 = 1.0003
        self.n2 = refractive_index
        self.water_roots_eps = water_roots_eps


def view_points_in_water(reconstructor, cam_id, pts3d, water, distorted=True):
    """
    pts3d : (N,3) array of 3D points

    returns:
      (2,N) projection of 3D points

    """
    assert isinstance(water, WaterInterface)

    pts3d = np.array(pts3d)
    assert pts3d.ndim == 2
    assert pts3d.shape[1] == 3  # pts3d.shape[0] == n_points
    n_points = pts3d.shape[0]

    pt1 = reconstructor.get_camera_center(cam_id)[:, 0]
    pt2 = np.array(pt1, copy=True)
    pt2[2] = 0  # closest point to camera on water surface, assumes water at z==0

    shifted_pts = pts3d - pt2  # origin under cam at surface. cam at (0,0,z).
    theta = np.arctan2(shifted_pts[:, 1], shifted_pts[:, 0])  # angles to points
    r = np.sqrt(np.sum(shifted_pts[:, :2] ** 2, axis=1))  # horizontal dist
    depth = -shifted_pts[:, 2]

    height = pt1[2]

    shifted_water_surface_pts = np.zeros_like(pts3d)

    r0 = []
    for i in range(len(shifted_pts)):
        assert depth[i] >= 0
        r0.append(
            _refraction.find_fastest_path_fermat(
                water.n1, water.n2, height, r[i], depth[i], water.water_roots_eps
            )
        )

    r0 = np.array(r0)
    shifted_water_surface_pts[:, 0] = r0 * np.cos(theta)
    shifted_water_surface_pts[:, 1] = r0 * np.sin(theta)

    water_surface_pts = np.ones((shifted_water_surface_pts.shape[0], 4))
    water_surface_pts[:, :3] = shifted_water_surface_pts + pt2

    pts = reconstructor.find2d(
        cam_id, water_surface_pts, distorted=distorted, bypass_refraction=True,
    )
    assert pts.shape == (2, n_points)
    return pts



import flydra_core._refraction as _refraction
import flydra_core._Roots3And4 as _Roots3And4

import numpy as np
import scipy.optimize
import math


def test_refraction():
    n1 = 1.0
    n2 = 2.0

    v1 = 1.0 / n1
    v2 = 1.0 / n2

    height = 1.0
    r = 1.2345

    # Test case where depth is zero: r == r0
    depth = 0.0

    eps = 1e-7

    r0 = _refraction.find_fastest_path_fermat(n1, n2, height, r, depth, eps)
    assert abs(r - r0) < 1e-6

    # Test case with real depth
    depth = 3.0
    r0 = _refraction.find_fastest_path_fermat(n1, n2, height, r, depth, eps)

    # test correctness with Snell's Law
    theta1 = np.arctan(r0 / height)
    theta2 = np.arctan((r - r0) / depth)

    ratio1 = v1 / v2
    ratio2 = np.sin(theta1) / np.sin(theta2)

    assert abs(ratio1 - ratio2) < 1e-6

    # test case that was actually failing in real-world use
    n1 = 1.0003
    n2 = 1.333
    height = 0.40279482775695
    r = 0.793324332905365
    depth = 1.7437460383163346
    r0 = _refraction.find_fastest_path_fermat(n1, n2, height, r, depth, eps)


def find_fastest_path_fermat_numeric(n1, n2, z1, h, z2):
    def duration(h1):
        # Fermat's principle of least time
        h2 = h - h1
        return n1 * math.sqrt(h1 * h1 + z1 * z1) + n2 * math.sqrt(z2 * z2 + h2 * h2)

    initial_r0 = h
    (final_r0,) = scipy.optimize.fmin(duration, [initial_r0], disp=0)
    return final_r0


def test_numeric_vs_algebraic():
    n1 = 1.0003
    n2 = 1.333
    height = 0.303
    r = 0.153
    depth = 0.010
    r0 = _refraction.find_fastest_path_fermat(n1, n2, height, r, depth, eps=1e-7)
    r0_numeric = find_fastest_path_fermat_numeric(n1, n2, height, r, depth)
    assert abs(r0 - r0_numeric) < 1e-5


def test_roots():
    poly = [
        -0.77628891,
        0.23754440646,
        -0.181206489286,
        0.0499191270735,
        -0.00381881322112,
    ]
    actual = _Roots3And4.py_real_nonnegative_root_less_than(
        poly[0], poly[1], poly[2], poly[3], poly[4], 0.153, 1e-7
    )
    expected_list = np.roots(poly)
    found = 0
    for maybe_expected in expected_list:
        if abs(actual - maybe_expected) < 1e-7:
            found += 1
    assert found == 1


if __name__ == "__main__":
    test_refraction()
    test_numeric_vs_algebraic()
    test_roots()



from __future__ import with_statement, division
import threading, time, socket, select, os, copy, struct
import warnings
import collections
import errno

import flydra_core.reconstruct
import flydra_core._reconstruct_utils as ru
import numpy
import numpy as np
from numpy import nan, inf
from flydra_core.common_variables import near_inf
import flydra_core.flydra_socket as flydra_socket
import Queue

pytables_filt = numpy.asarray
import pickle

import flydra_core.version
import flydra_core.kalman.flydra_kalman_utils as flydra_kalman_utils
import flydra_core.kalman.flydra_tracker
import flydra_core._fastgeom as geom

# import flydra_core.geom as geom

import flydra_core.data_descriptions

# ensure that pytables uses numpy:
import tables
import tables.flavor

tables.flavor.restrict_flavors(keep=["numpy"])
warnings.filterwarnings("ignore", category=tables.NaturalNameWarning)

import roslib

roslib.load_manifest("rospy")
roslib.load_manifest("ros_flydra")
import rospy
import std_msgs.msg
from ros_flydra.msg import flydra_mainbrain_super_packet, CameraList
from ros_flydra.msg import flydra_mainbrain_packet, flydra_object, FlydraError
from geometry_msgs.msg import Point, Vector3

import flydra_core.rosutils

LOG = flydra_core.rosutils.Log(to_ros=True)

import flydra_core.debuglock

DebugLock = flydra_core.debuglock.DebugLock

ATTEMPT_DATA_RECOVERY = bool(int(os.environ.get("ATTEMPT_DATA_RECOVERY", "1")))

IMPOSSIBLE_TIMESTAMP = -10.0

PT_TUPLE_IDX_X = flydra_core.data_descriptions.PT_TUPLE_IDX_X
PT_TUPLE_IDX_Y = flydra_core.data_descriptions.PT_TUPLE_IDX_Y
PT_TUPLE_IDX_FRAME_PT_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_FRAME_PT_IDX
PT_TUPLE_IDX_CUR_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_CUR_VAL_IDX
PT_TUPLE_IDX_MEAN_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_MEAN_VAL_IDX
PT_TUPLE_IDX_SUMSQF_VAL_IDX = flydra_core.data_descriptions.PT_TUPLE_IDX_SUMSQF_VAL_IDX


class CoordinateProcessor(threading.Thread):
    def __init__(
        self,
        main_brain,
        save_profiling_data,
        debug_level,
        show_sync_errors,
        show_overall_latency,
        max_reconstruction_latency_sec,
        max_N_hypothesis_test,
        use_unix_domain_sockets,
        posix_scheduler="",
    ):
        self.did_quit_successfully = False
        self.main_brain = main_brain
        self.debug_level = debug_level
        self.show_overall_latency = show_overall_latency
        self.max_reconstruction_latency_sec = max_reconstruction_latency_sec
        self.max_N_hypothesis_test = max_N_hypothesis_test
        self.posix_scheduler = posix_scheduler

        self._synchronized_cameras = []
        self.sync_cam_pub = None
        self.frame_offset_pubs = {}
        if rospy.core.is_initialized():
            self.sync_cam_pub = rospy.Publisher(
                "~synchronized_cameras", CameraList, queue_size=1, latch=True
            )

        self.save_profiling_data = save_profiling_data
        if self.save_profiling_data:
            self.data_dict_queue = []

        self.cam_ids = []
        self.absolute_cam_nos = []  # a.k.a. "camn"
        self.frame_offsets = {}
        self.last_frame_times = {}
        self.last_timestamps = []
        self.last_framenumbers_delay = []
        self.last_framenumbers_skip = []
        if ATTEMPT_DATA_RECOVERY:
            # self.request_data_lock = DebugLock('request_data_lock',True) # protect request_data
            self.request_data_lock = threading.Lock()  # protect request_data
            self.request_data = {}
        self.cam_id2cam_no = {}
        self.camn2cam_id = {}
        self.reconstructor = None
        self.tracker = None
        self.ever_synchronized = False
        self.show_sync_errors = show_sync_errors

        self.ip2hostname = {}

        self.tracker_lock = threading.Lock()
        # self.tracker_lock = DebugLock('tracker_lock',verbose=True)

        self.all_data_lock = threading.Lock()
        # self.all_data_lock = DebugLock('all_data_lock',verbose=False)
        self.quit_event = threading.Event()

        self.max_absolute_cam_nos = -1

        self.general_save_info = {}

        self.to_unlink = []

        if use_unix_domain_sockets:
            addr = "/tmp/flydra_coordinate_receiver." + str(os.getpid())
            try:
                os.remove(addr)
            except OSError as err:
                if err.errno != errno.ENOENT:  # Missing file is OK.
                    raise
            addrinfo = flydra_socket.make_addrinfo(filename=addr)
            self.to_unlink.append(addr)
        else:
            addrinfo = flydra_socket.make_addrinfo(
                host=flydra_socket.get_bind_address(), port=0
            )
        self.listen_socket = flydra_socket.FlydraTransportReceiver(
            addrinfo, socket_timeout_seconds=0.5
        )
        self.listen_address = self.listen_socket.get_listen_addrinfo().to_dict()
        if "host" in self.listen_address and self.listen_address["host"] == "0.0.0.0":
            self.listen_address["host"] = flydra_core.rosutils.get_node_hostname(
                rospy.get_name()
            )
        LOG.info("coordinate receiver listening at %r" % self.listen_address)

        self.set_reconstructor(None)

        self.realtime_ros_packets_pub = rospy.Publisher(
            "~super_packets",
            flydra_mainbrain_super_packet,
            queue_size=100,
            tcp_nodelay=True,
        )

        self.synchronze_ros_msgs_pub = rospy.Publisher(
            "~synchronize", std_msgs.msg.String, queue_size=100
        )

        self.realtime_coord_dict = {}
        self.timestamp_check_dict = {}
        self.realtime_kalman_coord_dict = collections.defaultdict(dict)
        self.oldest_timestamp_by_corrected_framenumber = {}

        threading.Thread.__init__(self, name="CoordinateProcessor")

    def get_listen_address(self):
        return self.listen_address

    def mainbrain_is_attempting_synchronizing(self):
        self.ever_synchronized = True
        with self.tracker_lock:
            if self.tracker is not None:
                self.tracker.kill_all_trackers()  # delete all old data

    def get_general_cam_info(self):
        with self.all_data_lock:
            result = self.general_save_info.copy()
        return result

    def get_missing_data_dict(self):
        # called from main thread, must lock data in realtime coord thread
        result_by_camn = {}
        with self.request_data_lock:
            for absolute_cam_no, tmp_queue in self.request_data.iteritems():
                list_of_missing_framenumbers = []
                cam_id = None
                framenumber_offset = None
                try:
                    while 1:
                        value = tmp_queue.get_nowait()
                        this_cam_id, this_framenumber_offset, this_list = value
                        if cam_id is None:
                            cam_id = this_cam_id
                        if framenumber_offset is None:
                            framenumber_offset = this_framenumber_offset

                        assert (
                            cam_id == this_cam_id
                        )  # make sure given camn comes from single cam_id
                        assert framenumber_offset == this_framenumber_offset

                        list_of_missing_framenumbers.extend(this_list)
                except Queue.Empty:
                    pass
                if len(list_of_missing_framenumbers):
                    result_by_camn[absolute_cam_no] = (
                        cam_id,
                        framenumber_offset,
                        list_of_missing_framenumbers,
                    )
        return result_by_camn

    def set_reconstructor(self, r):
        # called from main thread, must lock to send to realtime coord thread
        with self.all_data_lock:
            self.reconstructor = r

        R = self.reconstructor
        self.cached_calibration_by_cam_id = {}
        if R is None:
            return
        for cam_id in R.cam_ids:
            scc = R.get_SingleCameraCalibration(cam_id)
            cc = R.get_camera_center(cam_id)[:, 0]
            cc = np.array([cc[0], cc[1], cc[2], 1.0])
            self.cached_calibration_by_cam_id[cam_id] = scc, cc

    def set_new_tracker(self, kalman_model=None):
        # called from main thread, must lock to send to realtime coord thread
        tracker = flydra_core.kalman.flydra_tracker.Tracker(
            self.reconstructor, kalman_model=kalman_model
        )
        tracker.set_killed_tracker_callback(self.enqueue_finished_tracked_object)
        with self.tracker_lock:
            if self.tracker is not None:
                self.tracker.kill_all_trackers()  # save (if necessary) all old data
            self.tracker = tracker  # bind to name, replacing old tracker
            if self.save_profiling_data:
                tracker = copy.copy(self.tracker)
                tracker.kill_tracker_callbacks = []
                if 1:
                    raise NotImplementedError("")
                self.data_dict_queue.append(("tracker", tracker))

    def enqueue_finished_tracked_object(self, tracked_object):
        # this is from called within the realtime coords thread
        if self.main_brain.is_saving_data():
            self.main_brain.queue_data3d_kalman_estimates.put(
                (
                    tracked_object.obj_id,
                    tracked_object.frames,
                    tracked_object.xhats,
                    tracked_object.Ps,
                    tracked_object.timestamps,
                    tracked_object.observations_frames,
                    tracked_object.MLE_position,
                    tracked_object.observations_2d,
                    tracked_object.MLE_Lcoords,
                )
            )
        # send a ROS message that this object is dead
        this_ros_object = flydra_object(
            obj_id=tracked_object.obj_id,
            position=Point(numpy.nan, numpy.nan, numpy.nan),
        )
        try:
            reconstruction_stamp = rospy.Time.now()
            acquire_stamp = rospy.Time.from_sec(0)
        except rospy.exceptions.ROSInitException as err:
            pass
        else:
            ros_packet = flydra_mainbrain_packet(
                framenumber=tracked_object.current_frameno,
                reconstruction_stamp=reconstruction_stamp,
                acquire_stamp=acquire_stamp,
                objects=[this_ros_object],
            )
            self.realtime_ros_packets_pub.publish(
                flydra_mainbrain_super_packet([ros_packet])
            )

        if self.debug_level.isSet():
            LOG.debug("killing obj_id %d:" % tracked_object.obj_id)

    def connect(self, cam_id):
        # called from Remote-API thread on camera connect
        assert not self.main_brain.is_saving_data()

        with self.all_data_lock:
            self.cam_ids.append(cam_id)

            # find absolute_cam_no
            self.max_absolute_cam_nos += 1
            absolute_cam_no = self.max_absolute_cam_nos
            self.absolute_cam_nos.append(absolute_cam_no)

            self.camn2cam_id[absolute_cam_no] = cam_id
            self.cam_id2cam_no[cam_id] = absolute_cam_no

            self.last_timestamps.append(
                IMPOSSIBLE_TIMESTAMP
            )  # arbitrary impossible number
            self.last_framenumbers_delay.append(-1)  # arbitrary impossible number
            self.last_framenumbers_skip.append(-1)  # arbitrary impossible number
            self.general_save_info[cam_id] = {"absolute_cam_no": absolute_cam_no}

    def disconnect(self, cam_id):
        # called from Remote-API thread on camera disconnect
        cam_idx = self.cam_ids.index(cam_id)
        with self.all_data_lock:
            del self.cam_ids[cam_idx]
            del self.absolute_cam_nos[cam_idx]
            del self.last_timestamps[cam_idx]
            del self.last_framenumbers_delay[cam_idx]
            del self.last_framenumbers_skip[cam_idx]
            del self.general_save_info[cam_id]

    def quit(self):
        # called from outside of thread to quit the thread
        if self.save_profiling_data:
            fname = "data_for_kalman_profiling.pkl"
            fullpath = os.path.abspath(fname)
            LOG.info("saving data for profiling to %s" % fullpath)
            to_save = self.data_dict_queue
            save_fd = open(fullpath, mode="wb")
            pickle.dump(to_save, save_fd)
            save_fd.close()
            LOG.info("done saving")
        self.quit_event.set()
        self.join()  # wait until CoordReveiver thread quits

    def OnSynchronize(self, cam_idx, cam_id, framenumber, timestamp):

        if self.main_brain.is_saving_data():
            LOG.warn("re-synchronized while saving data!")
            return

        if 1:
            self.synchronze_ros_msgs_pub.publish(std_msgs.msg.String(cam_id))
            LOG.info("%s (re)synchronized" % cam_id)
            # discard all previous data
            for k in self.realtime_coord_dict.keys():
                del self.realtime_coord_dict[k]
                del self.timestamp_check_dict[k]
            for k in self.realtime_kalman_coord_dict.keys():
                del self.realtime_kalman_coord_dict[k]
            for k in self.oldest_timestamp_by_corrected_framenumber.keys():
                del self.oldest_timestamp_by_corrected_framenumber[k]

        # make new absolute_cam_no to indicate new synchronization state
        self.max_absolute_cam_nos += 1
        absolute_cam_no = self.max_absolute_cam_nos
        self.absolute_cam_nos[cam_idx] = absolute_cam_no

        self.camn2cam_id[absolute_cam_no] = cam_id
        self.cam_id2cam_no[cam_id] = absolute_cam_no

        self.general_save_info[cam_id]["absolute_cam_no"] = absolute_cam_no

        # because saving usually follows synchronization, ensure we have a recent
        # image to put in the h5 file
        self.main_brain.remote_api.external_request_image_async(cam_id)

    def _publish_list_of_synced_cameras(self):
        if self.sync_cam_pub is None:
            return
        msg = CameraList()
        msg.cameras = self._synchronized_cameras
        self.sync_cam_pub.publish(msg)

    def delete_list_of_synced_cameras(self):
        del self._synchronized_cameras[:]
        self._publish_list_of_synced_cameras()

    def _new_frame_offset(self, cam_id, framenumber):
        self.frame_offsets[cam_id] = framenumber
        if not rospy.core.is_initialized():
            return

        msg = std_msgs.msg.UInt64(data=framenumber)

        if cam_id not in self.frame_offset_pubs:
            pub = rospy.Publisher("/%s/frame_offset" % (cam_id,), std_msgs.msg.UInt64,)
            self.frame_offset_pubs[cam_id] = pub
        self.frame_offset_pubs[cam_id].publish(msg)

    def register_frame(self, cam_id, framenumber):
        frame_timestamp = time.time()
        try:
            last_frame_timestamp = self.last_frame_times[cam_id]
        except KeyError:
            # very first frame from this cam_id
            self._new_frame_offset(cam_id, framenumber)
            did_frame_offset_change = True
            corrected_framenumber = framenumber - self.frame_offsets[cam_id]
            self.last_frame_times[cam_id] = frame_timestamp
            return corrected_framenumber, did_frame_offset_change

        this_interval = frame_timestamp - last_frame_timestamp

        did_frame_offset_change = False
        if self.main_brain._is_synchronizing:
            if this_interval > flydra_core.common_variables.sync_duration:
                self._new_frame_offset(cam_id, framenumber)
                did_frame_offset_change = True
                self._synchronized_cameras.append(cam_id)
                self._publish_list_of_synced_cameras()

                if len(self._synchronized_cameras) == len(self.cam_ids):
                    # success, done synchronizing all cameras
                    self.main_brain._is_synchronizing = False
        else:
            if this_interval > flydra_core.common_variables.sync_duration:
                LOG.warn("long IFI not during intended synchronization detected")

        self.last_frame_times[cam_id] = frame_timestamp

        corrected_framenumber = framenumber - self.frame_offsets[cam_id]
        return corrected_framenumber, did_frame_offset_change

    def run(self):
        """main loop of CoordinateProcessor"""

        if self.posix_scheduler != "":
            import posix_sched

            mode_str, priority = self.posix_scheduler
            mode = getattr(posix_sched, mode_str)
            try:
                sched_params = posix_sched.SchedParam(priority)
                posix_sched.setscheduler(0, mode, sched_params)
                LOG.info(
                    "3D reconstruction thread running with priority %s"
                    % self.posix_scheduler
                )
            except Exception as x:
                LOG.warn(
                    "could not adjust priority (PID %d): %s" % (os.getpid(), str(x))
                )
        else:
            LOG.info("3D reconstruction thread running with default priority")
        LOG.info("CoordinateProcessor running on PID %d" % os.getpid())
        # self.main_brain.trigger_device.wait_for_estimate()
        while not self.quit_event.isSet():
            try:
                incoming_2d_data = self.listen_socket.recv()
            except socket.error as err:
                if err.errno == errno.EAGAIN:
                    # no data ready. try again (after checking if we should quit).
                    continue
                else:
                    raise
            self.process_data(incoming_2d_data)
        self.finish_processing()

    def finish_processing(self):
        with self.tracker_lock:
            if self.tracker is not None:
                self.tracker.kill_all_trackers()  # save (if necessary) all old data

        for fname in self.to_unlink:
            os.remove(fname)

        self.did_quit_successfully = True

    def process_data(self, buf_data):

        header_fmt = flydra_core.common_variables.recv_pt_header_fmt
        header_size = struct.calcsize(header_fmt)
        pt_fmt = flydra_core.common_variables.recv_pt_fmt
        pt_size = struct.calcsize(pt_fmt)

        no_point_tuple = (
            nan,
            nan,
            nan,
            nan,
            nan,
            nan,
            nan,
            nan,
            nan,
            False,
            0,
            0,
            0,
            0,
        )

        new_data_framenumbers = set()

        deferred_2d_data = []
        header = buf_data[:header_size]
        assert len(header) == header_size
        # this raw_timestamp is the remote camera's timestamp (?? from the driver, not the host clock??)
        (
            cam_id,
            raw_timestamp,
            camn_received_time,
            raw_framenumber,
            n_pts,
            n_frames_skipped,
        ) = struct.unpack(header_fmt, header)

        with self.all_data_lock:
            cam_idx = self.cam_ids.index(cam_id)
            absolute_cam_no = self.absolute_cam_nos[cam_idx]

            points_in_pluecker_coords_meters = []
            points_undistorted = []
            points_distorted = []

            assert len(buf_data) == (header_size + n_pts * pt_size)

            predicted_framenumber = (
                n_frames_skipped + self.last_framenumbers_skip[cam_idx] + 1
            )
            if raw_framenumber < predicted_framenumber:
                LOG.fatal("cam_id %s" % cam_id)
                LOG.fatal("raw_framenumber %s" % raw_framenumber)
                LOG.fatal("n_frames_skipped %s" % n_frames_skipped)
                LOG.fatal("predicted_framenumber %s" % predicted_framenumber)
                LOG.fatal(
                    "self.last_framenumbers_skip[cam_idx] %s"
                    % self.last_framenumbers_skip[cam_idx]
                )
                raise RuntimeError("got framenumber already received or skipped!")
            elif raw_framenumber > predicted_framenumber:
                if not self.last_framenumbers_skip[cam_idx] == -1:
                    # this is not the first frame
                    # probably because network buffer filled up before we emptied it
                    LOG.warn("frame data loss %s" % cam_id)
                    self.main_brain.error_ros_msgs_pub.publish(
                        FlydraError(FlydraError.FRAME_DATA_LOSS, cam_id)
                    )

                if ATTEMPT_DATA_RECOVERY:
                    if not self.last_framenumbers_skip[cam_idx] == -1:
                        # this is not the first frame
                        missing_frame_numbers = range(
                            self.last_framenumbers_skip[cam_idx] + 1, raw_framenumber
                        )

                        with self.request_data_lock:
                            tmp_queue = self.request_data.setdefault(
                                absolute_cam_no, Queue.Queue()
                            )

                        tmp_framenumber_offset = self.frame_offsets[cam_id]
                        tmp_queue.put(
                            (cam_id, tmp_framenumber_offset, missing_frame_numbers)
                        )
                        del tmp_framenumber_offset
                        del tmp_queue  # drop reference to queue
                        del missing_frame_numbers

            self.last_framenumbers_skip[cam_idx] = raw_framenumber
            start = header_size
            if n_pts:
                # valid points
                for frame_pt_idx in range(n_pts):
                    end = start + pt_size
                    (
                        x_distorted,
                        y_distorted,
                        area,
                        slope,
                        eccentricity,
                        slope_found,
                        cur_val,
                        mean_val,
                        sumsqf_val,
                    ) = struct.unpack(pt_fmt, buf_data[start:end])
                    # nan cannot get sent across network in platform-independent way

                    if slope == near_inf:
                        slope = inf
                    if eccentricity == near_inf:
                        eccentricity = inf
                    if not slope_found:
                        slope = nan

                    line_found = slope_found

                    if slope_found:
                        if np.isinf(slope):
                            run = 0.0
                            rise = 1.0
                        else:
                            run = 1.0
                            rise = slope
                    else:
                        run = nan
                        rise = nan

                    if cam_id in self.cached_calibration_by_cam_id:
                        scc, cc = self.cached_calibration_by_cam_id[cam_id]
                        x_undistorted, y_undistorted = scc.helper.undistort(
                            x_distorted, y_distorted
                        )
                        (
                            p1,
                            p2,
                            p3,
                            p4,
                            ray0,
                            ray1,
                            ray2,
                            ray3,
                            ray4,
                            ray5,
                        ) = flydra_core.reconstruct.do_3d_operations_on_2d_point(
                            scc.helper,
                            x_undistorted,
                            y_undistorted,
                            scc.pmat_inv,
                            cc,
                            x_distorted,
                            y_distorted,
                            rise,
                            run,
                        )
                        ray_valid = not np.isnan(ray0)
                    else:
                        scc, cc = None, None
                        x_undistorted, y_undistorted = np.nan, np.nan
                        (p1, p2, p3, p4, ray0, ray1, ray2, ray3, ray4, ray5) = [
                            np.nan
                        ] * 10
                        ray_valid = False

                    # Keep in sync with flydra_analysis.kalmanize and flydra_core.data_descriptions
                    pt_undistorted = (
                        x_undistorted,
                        y_undistorted,
                        area,
                        slope,
                        eccentricity,
                        p1,
                        p2,
                        p3,
                        p4,
                        line_found,
                        frame_pt_idx,
                        cur_val,
                        mean_val,
                        sumsqf_val,
                    )
                    pt_distorted = (
                        x_distorted,
                        y_distorted,
                        area,
                        slope,
                        eccentricity,
                        p1,
                        p2,
                        p3,
                        p4,
                        line_found,
                        frame_pt_idx,
                        cur_val,
                        mean_val,
                        sumsqf_val,
                    )
                    if ray_valid:
                        points_in_pluecker_coords_meters.append(
                            (
                                pt_undistorted,
                                geom.line_from_HZline(
                                    (ray0, ray1, ray2, ray3, ray4, ray5)
                                ),
                            )
                        )
                    points_undistorted.append(pt_undistorted)
                    points_distorted.append(pt_distorted)
                    start = end
            else:
                # no points found
                end = start
                # append non-point to allow correlation of
                # timestamps with frame number
                points_distorted.append(no_point_tuple)
                points_undistorted.append(no_point_tuple)

            self._process_parsed_data(
                cam_idx,
                camn_received_time,
                absolute_cam_no,
                n_pts,
                cam_id,
                raw_framenumber,
                new_data_framenumbers,
                points_in_pluecker_coords_meters,
                points_distorted,
                points_undistorted,
                deferred_2d_data,
            )

    def _process_parsed_data(
        self,
        cam_idx,
        camn_received_time,
        absolute_cam_no,
        n_pts,
        cam_id,
        raw_framenumber,
        new_data_framenumbers,
        points_in_pluecker_coords_meters,
        points_distorted,
        points_undistorted,
        deferred_2d_data,
    ):
        # Note: this must be called with self.all_data_lock acquired.

        timestamp_check_dict = self.timestamp_check_dict
        convert_format = flydra_kalman_utils.convert_format  # shorthand
        realtime_coord_dict = self.realtime_coord_dict
        oldest_timestamp_by_corrected_framenumber = (
            self.oldest_timestamp_by_corrected_framenumber
        )
        realtime_kalman_coord_dict = self.realtime_kalman_coord_dict

        if 1:
            if 1:
                if 1:
                    if 1:
                        # Use camn_received_time to determine sync
                        # info. This avoids 2 potential problems:
                        #  * using raw_timestamp can fail if the
                        #    camera drivers don't provide useful data
                        #  * using time.time() can fail if the network
                        #    latency jitter is on the order of the
                        #    inter frame interval.
                        (
                            corrected_framenumber,
                            did_frame_offset_change,
                        ) = self.register_frame(cam_id, raw_framenumber)

                        trigger_timestamp = self.main_brain.trigger_device.framestamp2timestamp(
                            corrected_framenumber
                        )
                        if did_frame_offset_change:
                            self.OnSynchronize(
                                cam_idx, cam_id, raw_framenumber, trigger_timestamp
                            )
                            new_data_framenumbers.clear()

                        self.last_timestamps[cam_idx] = trigger_timestamp
                        self.last_framenumbers_delay[cam_idx] = raw_framenumber
                        self.main_brain.framenumber = corrected_framenumber

                        if self.main_brain.is_saving_data():
                            for point_tuple in points_distorted:
                                # Save 2D data (even when no point found) to allow
                                # temporal correlation of movie frames to 2D data.
                                frame_pt_idx = point_tuple[PT_TUPLE_IDX_FRAME_PT_IDX]
                                cur_val = point_tuple[PT_TUPLE_IDX_CUR_VAL_IDX]
                                mean_val = point_tuple[PT_TUPLE_IDX_MEAN_VAL_IDX]
                                sumsqf_val = point_tuple[PT_TUPLE_IDX_SUMSQF_VAL_IDX]
                                if corrected_framenumber is None:
                                    # don't bother saving if we don't know when it was from
                                    continue
                                deferred_2d_data.append(
                                    (
                                        absolute_cam_no,  # defer saving to later
                                        corrected_framenumber,
                                        trigger_timestamp,
                                        camn_received_time,
                                    )
                                    + point_tuple[:5]
                                    + (frame_pt_idx, cur_val, mean_val, sumsqf_val)
                                )
                        # save new frame data

                        if corrected_framenumber not in realtime_coord_dict:
                            realtime_coord_dict[corrected_framenumber] = {}
                            timestamp_check_dict[corrected_framenumber] = {}

                        # For hypothesis testing: attempt 3D reconstruction of 1st point from each 2D view
                        realtime_coord_dict[corrected_framenumber][
                            cam_id
                        ] = points_undistorted[0]
                        # timestamp_check_dict[corrected_framenumber][cam_id]= camn_received_time
                        timestamp_check_dict[corrected_framenumber][
                            cam_id
                        ] = trigger_timestamp

                        if len(points_in_pluecker_coords_meters):
                            # save all 3D Pluecker coordinates for Kalman filtering
                            realtime_kalman_coord_dict[corrected_framenumber][
                                absolute_cam_no
                            ] = points_in_pluecker_coords_meters

                        if n_pts:
                            inc_val = 1
                        else:
                            inc_val = 0

                        if (
                            corrected_framenumber
                            in oldest_timestamp_by_corrected_framenumber
                        ):
                            (
                                orig_timestamp,
                                n,
                            ) = oldest_timestamp_by_corrected_framenumber[
                                corrected_framenumber
                            ]
                            if orig_timestamp is None:
                                oldest = trigger_timestamp  # this may also be None, but eventually won't be
                            else:
                                oldest = min(trigger_timestamp, orig_timestamp)
                            oldest_timestamp_by_corrected_framenumber[
                                corrected_framenumber
                            ] = (oldest, n + inc_val)
                            del oldest, n, orig_timestamp
                        else:
                            oldest_timestamp_by_corrected_framenumber[
                                corrected_framenumber
                            ] = (trigger_timestamp, inc_val)

                        new_data_framenumbers.add(
                            corrected_framenumber
                        )  # insert into set

                finished_corrected_framenumbers = []  # for quick deletion

                ########################################################################

                # Now we've grabbed all data waiting on network. Now it's
                # time to calculate 3D info.

                # XXX could go for latest data first to minimize latency
                # on that data.

                ########################################################################

                for corrected_framenumber in new_data_framenumbers:
                    (
                        oldest_camera_timestamp,
                        n,
                    ) = oldest_timestamp_by_corrected_framenumber[corrected_framenumber]
                    if oldest_camera_timestamp is None:
                        # LOG.info('no latency estimate available -- skipping 3D reconstruction')
                        continue
                    if (
                        time.time() - oldest_camera_timestamp
                    ) > self.max_reconstruction_latency_sec:
                        # LOG.info('maximum reconstruction latency exceeded -- skipping 3D reconstruction')
                        continue

                    data_dict = realtime_coord_dict[corrected_framenumber]
                    if len(data_dict) == len(self.cam_ids):  # all camera data arrived

                        if self.debug_level.isSet():
                            LOG.debug("frame %d" % (corrected_framenumber))

                        # mark for deletion out of data queue
                        finished_corrected_framenumbers.append(corrected_framenumber)

                        if self.reconstructor is None:
                            # can't do any 3D math without calibration information
                            self.main_brain.best_realtime_data = None
                            continue

                        if not self.main_brain.trigger_device.have_estimate():
                            # acquire_stamp (the proximate error, however latency estimates for the same reason)
                            # cannot be calculated unless the triggerbox has a clock model
                            continue

                        if 1:
                            with self.tracker_lock:
                                if (
                                    self.tracker is None
                                ):  # tracker isn't instantiated yet...
                                    self.main_brain.best_realtime_data = None
                                    continue

                                pluecker_coords_by_camn = realtime_kalman_coord_dict[
                                    corrected_framenumber
                                ]

                                if self.save_profiling_data:
                                    dumps = pickle.dumps(pluecker_coords_by_camn)
                                    self.data_dict_queue.append(
                                        (
                                            "gob",
                                            (
                                                corrected_framenumber,
                                                dumps,
                                                self.camn2cam_id,
                                            ),
                                        )
                                    )
                                pluecker_coords_by_camn = self.tracker.calculate_a_posteriori_estimates(
                                    corrected_framenumber,
                                    pluecker_coords_by_camn,
                                    self.camn2cam_id,
                                    debug2=self.debug_level.isSet(),
                                )

                                if self.debug_level.isSet():
                                    LOG.debug(
                                        "%d live objects:"
                                        % self.tracker.how_many_are_living()
                                    )
                                    results = self.tracker.get_most_recent_data()
                                    for result in results:
                                        if result is None:
                                            continue
                                        obj_id, last_xhat, P = result
                                        Pmean = numpy.sqrt(
                                            numpy.sum([P[i, i] ** 2 for i in range(3)])
                                        )
                                        LOG.debug(
                                            "obj_id %d: (%.3f, %.3f, %.3f), Pmean: %.3f"
                                            % (
                                                obj_id,
                                                last_xhat[0],
                                                last_xhat[1],
                                                last_xhat[2],
                                                Pmean,
                                            )
                                        )

                                if self.save_profiling_data:
                                    self.data_dict_queue.append(
                                        ("ntrack", self.tracker.how_many_are_living())
                                    )

                                now = time.time()
                                if self.show_overall_latency.isSet():
                                    (
                                        oldest_camera_timestamp,
                                        n,
                                    ) = oldest_timestamp_by_corrected_framenumber[
                                        corrected_framenumber
                                    ]
                                    if n > 0:
                                        if 0:
                                            LOG.info(
                                                "overall latency %d: %.1f msec (oldest: %s now: %s)"
                                                % (
                                                    n,
                                                    (now - oldest_camera_timestamp)
                                                    * 1e3,
                                                    repr(oldest_camera_timestamp),
                                                    repr(now),
                                                )
                                            )
                                        else:

                                            LOG.info(
                                                "overall latency (%d camera detected 2d points): %.1f msec (note: may exclude camera->camera computer latency)"
                                                % (
                                                    n,
                                                    (now - oldest_camera_timestamp)
                                                    * 1e3,
                                                )
                                            )

                                if 1:
                                    # The above calls
                                    # self.enqueue_finished_tracked_object()
                                    # when a tracked object is no longer
                                    # tracked.

                                    # Now, tracked objects have been updated (and their 2D data points
                                    # removed from consideration), so we can use old flydra
                                    # "hypothesis testing" algorithm on remaining data to see if there
                                    # are new objects.

                                    results = self.tracker.get_most_recent_data()
                                    Xs = []
                                    for result in results:
                                        if result is None:
                                            continue
                                        obj_id, last_xhat, P = result
                                        X = last_xhat[0], last_xhat[1], last_xhat[2]
                                        Xs.append(X)
                                    if len(Xs):
                                        self.main_brain.best_realtime_data = Xs, 0.0
                                    else:
                                        self.main_brain.best_realtime_data = None

                                # Convert to format accepted by find_best_3d()
                                found_data_dict, first_idx_by_cam_id = convert_format(
                                    pluecker_coords_by_camn,
                                    self.camn2cam_id,
                                    area_threshold=0.0,
                                    only_likely=True,
                                )

                                if len(found_data_dict) >= 2:
                                    # Can't do any 3D math without at least 2 cameras giving good
                                    # data.
                                    max_error = self.tracker.kalman_model[
                                        "hypothesis_test_max_acceptable_error"
                                    ]
                                    with_water = self.reconstructor.wateri is not None

                                    try:
                                        (
                                            this_observation_3d,
                                            this_observation_Lcoords,
                                            cam_ids_used,
                                            min_mean_dist,
                                        ) = ru.hypothesis_testing_algorithm__find_best_3d(
                                            self.reconstructor,
                                            found_data_dict,
                                            max_error,
                                            max_n_cams=self.max_N_hypothesis_test,
                                            with_water=with_water,
                                        )
                                    except ru.NoAcceptablePointFound:
                                        pass
                                    else:
                                        this_observation_camns = [
                                            self.cam_id2cam_no[cam_id]
                                            for cam_id in cam_ids_used
                                        ]
                                        this_observation_idxs = [
                                            first_idx_by_cam_id[cam_id]
                                            for cam_id in cam_ids_used
                                        ]
                                        ####################################
                                        #  Now join found point into Tracker
                                        if self.save_profiling_data:
                                            self.data_dict_queue.append(
                                                (
                                                    "join",
                                                    (
                                                        corrected_framenumber,
                                                        this_observation_3d,
                                                        this_observation_Lcoords,
                                                        this_observation_camns,
                                                        this_observation_idxs,
                                                    ),
                                                )
                                            )
                                        # test for novelty
                                        believably_new = self.tracker.is_believably_new(
                                            this_observation_3d
                                        )
                                        if believably_new:
                                            self.tracker.join_new_obj(
                                                corrected_framenumber,
                                                this_observation_3d,
                                                this_observation_Lcoords,
                                                this_observation_camns,
                                                this_observation_idxs,
                                            )
                                if 1:
                                    if self.tracker.how_many_are_living():
                                        # publish state to ROS
                                        results = self.tracker.get_most_recent_data()
                                        ros_objects = []
                                        for result in results:
                                            if result is None:
                                                continue
                                            obj_id, xhat, P = result
                                            this_ros_object = flydra_object(
                                                obj_id=obj_id,
                                                position=Point(*xhat[:3]),
                                                velocity=Vector3(*xhat[3:6]),
                                                posvel_covariance_diagonal=numpy.diag(
                                                    P
                                                )[:6].tolist(),
                                            )
                                            ros_objects.append(this_ros_object)
                                        ros_packet = flydra_mainbrain_packet(
                                            framenumber=corrected_framenumber,
                                            reconstruction_stamp=rospy.Time.from_sec(
                                                now
                                            ),
                                            acquire_stamp=rospy.Time.from_sec(
                                                oldest_camera_timestamp
                                            ),
                                            objects=ros_objects,
                                        )
                                        self.realtime_ros_packets_pub.publish(
                                            flydra_mainbrain_super_packet([ros_packet])
                                        )

                for finished in finished_corrected_framenumbers:
                    # check that timestamps are in reasonable agreement (low priority)
                    diff_from_start = []
                    for cam_id, tmp_trigger_timestamp in timestamp_check_dict[
                        finished
                    ].iteritems():
                        diff_from_start.append(tmp_trigger_timestamp)
                    timestamps_by_cam_id = numpy.array(diff_from_start)

                    if self.show_sync_errors:
                        if len(timestamps_by_cam_id):
                            if (
                                numpy.max(
                                    abs(timestamps_by_cam_id - timestamps_by_cam_id[0])
                                )
                                > 0.005
                            ):
                                LOG.warn(
                                    "timestamps off by more than 5 msec -- synchronization error"
                                )
                                self.main_brain.error_ros_msgs_pub.publish(
                                    FlydraError(FlydraError.CAM_TIMESTAMPS_OFF, "")
                                )

                    del realtime_coord_dict[finished]
                    del timestamp_check_dict[finished]
                    try:
                        del realtime_kalman_coord_dict[finished]
                    except KeyError:
                        pass

                # Clean up old frame records to save RAM.
                # This is only needed when multiple cameras are not
                # synchronized, (When camera-camera frame
                # correspondences are unknown.)
                if len(realtime_coord_dict) > 100:
                    # dont spam the console at startup (i.e. before a sync has been attemted)
                    if self.ever_synchronized:
                        LOG.warn(
                            "Cameras not synchronized or network dropping packets -- unmatched 2D data accumulating"
                        )
                        self.main_brain.error_ros_msgs_pub.publish(
                            FlydraError(FlydraError.NOT_SYNCHRONIZED, "")
                        )

                    k = realtime_coord_dict.keys()
                    k.sort()

                    # get one sample
                    corrected_framenumber = k[0]
                    data_dict = realtime_coord_dict[corrected_framenumber]
                    this_cam_ids = data_dict.keys()
                    missing_cam_id_guess = list(set(self.cam_ids) - set(this_cam_ids))
                    if len(missing_cam_id_guess) and self.ever_synchronized:
                        delta = list(set(self.cam_ids) - set(this_cam_ids))
                        LOG.warn("a guess at missing cam_id(s): %r" % delta)
                        for d in delta:
                            self.main_brain.error_ros_msgs_pub.publish(
                                FlydraError(FlydraError.MISSING_DATA, d)
                            )

                    for ki in k[:-50]:
                        del realtime_coord_dict[ki]
                        del timestamp_check_dict[ki]

                if len(realtime_kalman_coord_dict) > 100:
                    LOG.warn(
                        "deleting unused 3D data (this should be a rare occurrance)"
                    )
                    self.main_brain.error_ros_msgs_pub.publish(
                        FlydraError(FlydraError.UNUSED_3D_DATA, "")
                    )
                    k = realtime_kalman_coord_dict.keys()
                    k.sort()
                    for ki in k[:-50]:
                        del realtime_kalman_coord_dict[ki]

                if len(oldest_timestamp_by_corrected_framenumber) > 100:
                    k = oldest_timestamp_by_corrected_framenumber.keys()
                    k.sort()
                    for ki in k[:-50]:
                        del oldest_timestamp_by_corrected_framenumber[ki]

                if len(deferred_2d_data):
                    self.main_brain.queue_data2d.put(deferred_2d_data)
